function [B ax ay dBdv1 dBdv2 dBdv3] = blurImage(E, v, interp, dEdx, dEdy) 

if (nargin < 3) 
    interp = 'bilinear'; 
end

if (nargout < 4) 
    derivative = 0; 
else
    derivative = 1;
end

nT = size(v, 2); 
nR = size(E, 1); 
nC = size(E, 2); 
ridx = repmat([1:nR]', 1, nC);
cidx = repmat([1:nC], nR, 1);

mean_x = size(E,2)/2; 
mean_y = size(E,1)/2; 

% Shift the coordinate system so that the origin lies in the centre, and 
% y-axis goes up
rx = cidx - 0.5 - mean_x; 
ry = -((ridx - 0.5) - mean_y);

ax = zeros(nR, nC, nT+1); 
ay = ax;  
ax(:,:,1) = rx; 
ay(:,:,1) = ry;

p = zeros(3, nT+1);

% Numerically integrate the velocity to get the trajectories of each pixel
for t = 1:size(v,2)
    p(:,t+1) = v(:,t) + p(:,t);
    Vx = -[ay(:,:,t) - p(2,t)]*v(3,t) + v(1,t);
    Vy = [ax(:,:,t) - p(1,t)]*v(3,t) + v(2,t);
    ax(:,:,t+1) = ax(:,:,t) + Vx;
    ay(:,:,t+1) = ay(:,:,t) + Vy;
end

% Convert back to row-column space, pc and pr are stacks of column and 
% row coordinates through time
pc = ax + mean_x + 0.5;
pr = -ay + mean_y + 0.5; 

if (strcmp(interp, 'bilinear'))
    B = zeros(size(E));
    if (derivative == 1)
        dBdv1 = zeros(nR, nC, nT);
        dBdv2 = zeros(nR, nC, nT);
        dBdv3 = zeros(nR, nC, nT);
    end
    
    % We need the 4 pixels near by to do bilinear interpolation, alpha 
    % represents the mixing portion
    cpc = ceil(pc);
    cpr = ceil(pr);
    fpc = floor(pc);
    fpr = floor(pr);
    alphac = pc - fpc;
    alphar = pr - fpr;
    
    % Handle image boundary by clipping the coordinates to the edges
    fpr(fpr < 1) = 1; 
    fpr(fpr >= nR) = nR; 
    cpr(cpr < 1) = 1; 
    cpr(cpr >= nR) = nR; 
    fpc(fpc < 1) = 1; 
    fpc(fpc >= nC) = nC; 
    cpc(cpc < 1) = 1; 
    cpc(cpc >= nC) = nC; 
    
    i00 = nR*(fpc - 1) + fpr;
    i01 = nR*(cpc - 1) + fpr; 
    i10 = nR*(fpc - 1) + cpr;
    i11 = nR*(cpc - 1) + cpr; 

    for t = 1:nT+1
        E00 = E(i00(:,:,t)); 
        E01 = E(i01(:,:,t)); 
        E10 = E(i10(:,:,t)); 
        E11 = E(i11(:,:,t)); 
        
        % Do bilinear interpolation, and operate on the whole image every 
        % time step
        U = (1 - alphac(:,:,t)).*E00 + alphac(:,:,t).*E01; 
        L = (1 - alphac(:,:,t)).*E10 + alphac(:,:,t).*E11;
        
        B = B + (1 - alphar(:,:,t)).*U + alphar(:,:,t).*L; 
        
        if (derivative == 1 && t > 1)
            % Calculate the image derivative at time t
            dU = E01 - E00;
            dLo = E11 - E10;
            dLe = E10 - E00;
            dR = E11 - E01;
            dBdv1(:,:,t-1) = ((1-alphar(:,:,t)).*dU + alphar(:,:,t).*dLo);
            dBdv2(:,:,t-1) = -((1-alphac(:,:,t)).*dLe + alphac(:,:,t).*dR);
        end

% Alternatively, can do a double for-loop to do the same thing as above
%
%         row(:,:,1) = fpr(:,:,t);
%         row(:,:,2) = cpr(:,:,t);
%         col(:,:,1) = fpc(:,:,t);
%         col(:,:,2) = cpc(:,:,t);
%         
%         for r = 1:size(E,1)
%             for c = 1:size(E,2)
%                 Upper = (1-alphac(r,c,t))*E(row(r,c,1), col(r,c,1)) + ...
%                     alphac(r,c,t)*E(row(r,c,1), col(r,c,2));
%                 Lower = (1-alphac(r,c,t))*E(row(r,c,2), col(r,c,1)) + ...
%                     alphac(r,c,t)*E(row(r,c,2), col(r,c,2));
%                 B(r,c) = B(r, c) + (1-alphar(r,c,t))*Upper + alphar(r,c,t)*Lower;
% 
%                 if (derivative == 1)
%                     if (t > 1)
%                         dUpper = E(row(r,c,1), col(r,c,2)) - E(row(r,c,1), col(r,c,1));
%                         dLower = E(row(r,c,2), col(r,c,2)) - E(row(r,c,2), col(r,c,1));
%                         dLeft = E(row(r,c,2), col(r,c,1)) - E(row(r,c,1), col(r,c,1));
%                         dRight = E(row(r,c,2), col(r,c,2)) - E(row(r,c,1), col(r,c,2));
%                         dBdv1(r,c,t-1) = ((1-alphar(r,c,t))*dUpper + alphar(r,c,t)*dLower);
%                         dBdv2(r,c,t-1) = -((1-alphac(r,c,t))*dLeft + alphac(r,c,t)*dRight);
%                     end
%                 end
%             end
%         end
    end
    B = B ./ (nT+1); 
    
    if (derivative == 1)
        W = zeros(2,2,nT);
        W(:,:,1) = eye(2); 
        invW = zeros(2,2,nT);
        invW(:,:,1) = eye(2); 
        for t = 2:nT
            W(:,:,t) = W(:,:,t-1)*[1 -v(3,t); v(3, t) 1];
            invW(:,:,t) = inv([1 -v(3,t); v(3, t) 1])*invW(:,:,t-1); 
        end
        
                
        for t = 1:nT
            sum1 = zeros(nR, nC);
            sum2 = zeros(nR, nC);
            for tt = t:nT
                WW = invW(:,:,t)*W(:,:,tt);
                sum1 = sum1 + dBdv1(:,:,tt)*WW(1,1) + dBdv2(:,:,tt)*WW(2,1);
                sum2 = sum2 + dBdv1(:,:,tt)*WW(1,2) + dBdv2(:,:,tt)*WW(2,2);
            end
            dBdv3(:,:,t) = sum1.*(-[ay(:,:,t) - p(2,t)]) + sum2.*([ax(:,:,t) - p(1,t)]);
        end
        
        
%         daxdw = zeros(nR, nC, nT);
%         daydw = zeros(nR, nC, nT);
%         
%         % Calculate derivatives w.r.t. angular velocity, this is the
%         % bottleneck right now...there should be a faster way
%        
%  
%         for t = 1:nT
% 
%             for tt = t:nT
%                 if (tt == t)
%                     daxdw(:,:,tt) = -[ay(:,:,tt) - p(2,tt)];
%                     daydw(:,:,tt) = ax(:,:,tt) - p(1,tt);
%                 elseif (tt > t)
%                     daxdw(:,:,tt) = daxdw(:,:,tt-1) - daydw(:,:,tt-1)*v(3,tt);
%                     daydw(:,:,tt) = daydw(:,:,tt-1) + daxdw(:,:,tt-1)*v(3,tt);
%                 end
% 
%                 dBdv3(:,:,t) = dBdv3(:,:,t) + dBdv1(:,:,tt).*daxdw(:,:,tt) + dBdv2(:,:,tt).*daydw(:,:,tt);
%             end   
%         end
        

        % Calculate derivatives w.r.t. x and y velocity, it just happens
        % that daxdv and daydv are either 1 or 0...so dBdv1 and dBdv2 can
        % be computed this way.
        for t = nT:-1:2
            dBdv1(:,:,t-1) = dBdv1(:,:,t-1) + dBdv1(:,:,t);
            dBdv2(:,:,t-1) = dBdv2(:,:,t-1) + dBdv2(:,:,t);
        end


        dBdv1 = dBdv1 ./ (nT+1);
        dBdv2 = dBdv2 ./ (nT+1);
        dBdv3 = dBdv3 ./ (nT+1);
    end
end

% nearest neighbour doesn't work for gradient-based optimization
if (strcmp(interp, 'nearest'))
    iax = round(pc);
    iay = round(pr);


    B = zeros(nR, nC);
    if (derivative == 1)
        dBdv1 = zeros(nR, nC, nT);
        dBdv2 = zeros(nR, nC, nT);
        dBdv3 = zeros(nR, nC, nT);
    end
    for t = 1:nT+1
        row = iay(:,:,t);
        col = iax(:,:,t);
        row(row < 1) = 1;
        col(col < 1) = 1;
        row(row >= nR) = nR;
        col(col >= nC) = nC;

        for r = 1:size(E,1)
            for c = 1:size(E,2)
                B(r,c) = B(r, c) + E(row(r, c), col(r, c));
                if (derivative == 1)
                    if (t > 1)
                        dBdv1(r,c,t-1) = dEdx(row(r,c), col(r, c));
                        dBdv2(r,c,t-1) = -dEdy(row(r,c), col(r, c));
                    end
                end
            end
        end
    end
    B = B ./ (nT+1);

    if (derivative == 1)

        
        daxdw = zeros(nR, nC, nT);
        daydw = zeros(nR, nC, nT);
  
        for t = 1:nT
            for tt = t:nT
                if (tt == t)
                    daxdw(:,:,tt) = -[ay(:,:,tt) - p(2,tt)];
                    daydw(:,:,tt) = ax(:,:,tt) - p(1,tt);
                elseif (tt > t)
                    daxdw(:,:,tt) = daxdw(:,:,tt-1) - daydw(:,:,tt-1)*v(3,tt);
                    daydw(:,:,tt) = daydw(:,:,tt-1) + daxdw(:,:,tt-1)*v(3,tt);
                end

                dBdv3(:,:,t) = dBdv3(:,:,t) + dBdv1(:,:,tt).*daxdw(:,:,tt) + dBdv2(:,:,tt).*daydw(:,:,tt);
            end
        end
        


        for t = nT:-1:2
            dBdv1(:,:,t-1) = dBdv1(:,:,t-1) + dBdv1(:,:,t);
            dBdv2(:,:,t-1) = dBdv2(:,:,t-1) + dBdv2(:,:,t);
        end


        dBdv1 = dBdv1 ./ (nT+1);
        dBdv2 = dBdv2 ./ (nT+1);
        dBdv3 = dBdv3 ./ (nT+1);
    end
end

